"""
decorator objects for other functions in the query module
"""

from ..connection import rethinkdb as r, connect
from ..connection import validate_get_dbs
from decorator import decorator

WRAP_RETHINK_ERRORS = (r.errors.ReqlOpFailedError,
                       r.errors.ReqlError,
                       r.errors.ReqlDriverError)


@decorator
def wrap_rethink_errors(func_, *args, **kwargs):
    """
    Wraps rethinkdb specific errors as builtin/Brain errors

    :param func_: <function> to call
    :param args:  <tuple> positional arguments
    :param kwargs: <dict> keyword arguments
    :return: inherits from the called function
    """
    try:
        return func_(*args, **kwargs)
    except WRAP_RETHINK_ERRORS as e:
        raise ValueError(str(e))

@decorator
def wrap_rethink_generator_errors(func_, *args, **kwargs):
    """

    :param func_: <function> to call
    :param args:  <tuple> positional arguments
    :param kwargs: <dict> keyword arguments
    :return: inherits from the called function
    """
    try:
        for data in func_(*args, **kwargs):
            yield data
    except WRAP_RETHINK_ERRORS as e:
        raise ValueError(str(e))


@decorator
def wrap_connection_reconnect_test(func_, *args, **kwargs):
    """
    if a connection argument is passed, verify connection is good
    if not connected, attempt reconnect
    if reconnect fails, raises the rethink error
    other decorator will translate to standard error

    note: <rethinkdb.DefaultConnection>.is_open() appears to hang
            when the database is pulled out from under conn obj

    should be decorated prior to wrap_connection

    :param func_: <function> to call
    :param args:  <tuple> positional arguments
    :param kwargs: <dict> keyword arguments
    :return:
    :raises: ReqlDriverError or ConnectionRefusedError if reconnect failed.
    """
    conn = args[-1]
    if conn:  # conn object attempted
        try:
            validate_get_dbs(conn)
        except WRAP_RETHINK_ERRORS:
            conn.reconnect()  #throw may occur here
    return func_(*args, **kwargs)


@decorator
@wrap_connection_reconnect_test
def wrap_connection(func_, *args, **kwargs):
    """
    conn (connection) must be the last positional argument
    in all wrapped functions

    :param func_: <function> to call
    :param args:  <tuple> positional arguments
    :param kwargs: <dict> keyword arguments
    :return:
    """
    if not args[-1]:
        new_args = list(args)
        new_args[-1] = connect()
        args = tuple(new_args)
    return func_(*args, **kwargs)



"""
module functions to establish a brain connection
"""

from time import sleep, time
from uuid import uuid4
import rethinkdb
from rethinkdb.net import DefaultConnection
from decorator import decorator
from .environment import check_stage_env

#Recursive imports at bottom of file

BRAIN_DB = "Brain"
AUDIT_DB = "Audit"
PLUGINDB = "Plugins"
JOBS = "Jobs"
TARGETS = "Targets"
OUTPUTS = "Outputs"
SELF_TEST = {BRAIN_DB: [JOBS, TARGETS, OUTPUTS],
             #AUDIT_DB: [], #Audit DB soon into latest
             PLUGINDB: []}

DEFAULT_HOSTS = {"PROD": "rethinkdb",
                 "QA": "rethinkdb",
                 "DEV": "localhost",
                 "TESTING": "localhost",
                 "": "localhost"}  # environment not configured, try anyway


@decorator
def wrap_self_test(func_, *args, **kwargs):
    """
    should be applied as decorator to functions
    requiring a SELF_TEST dict

    If the caller does not supply a self_test dict, it supplies one

    :param f: <function>
    :param args: <tuple> positional arguments
    :param kwargs: <dict> keyword arguments
    :return: return value of the called function
    """
    if not args[-1]:
        new_args = list(args)
        new_args[-1] = SELF_TEST
        args = tuple(new_args)
    return func_(*args, **kwargs)


class BrainNotReady(Exception):
    """
    Simple exception to identify that the brain is not ready for use at this time
    """
    pass


def validate_get_dbs(connection):
    """
    validates the connection object is capable of read access to rethink

    should be at least one test database by default

    :param connection: <rethinkdb.net.DefaultConnection>
    :return: <set> list of databases
    :raises: ReqlDriverError AssertionError
    """
    remote_dbs = set(rethinkdb.db_list().run(connection))
    assert remote_dbs
    return remote_dbs


def validate_brain_requirements(connection, remote_dbs, requirements):
    """
    validates the rethinkdb has the 'correct' databases and tables
    should get remote_dbs from brain.connection.validate_get_dbs

    :param connection: <rethinkdb.net.DefaultConnection>
    :param remote_dbs: <set> database names present in remote database
    :param requirements: <dict> example(brain.connection.SELF_TEST)
    :return: <bool> True
    :raises: AssertionError or Reql*Error
    """
    for database in requirements:
        assert (database in remote_dbs), "database {} must exist".format(database)
        remote_tables = frozenset(rethinkdb.db(database).table_list().run(connection))
        for table in requirements[database]:
            assert (table in remote_tables), "{} must exist in {}".format(table, database)
    return True


def validate_write_access(connection):
    """
    verifies connection can write to rethink by:
        1. Creating a test table
        2. Adding a document to that table
        3. destroying the test table

    :param connection: <rethinkdb.net.DefaultConnection>
    :return: <bool> True
    :raises: Reql*Error
    """
    test_table = "test_table_{}".format(uuid4()).replace("-", "")   # '-' is not valid
    create_plugin(test_table, connection)
    RPX.table(test_table).insert({"test": "data"}).run(connection)
    destroy_plugin(test_table, connection)
    return True


def validate_brain(connection, requirements=None):
    """
    Alias to brain_post function

    Checks that the brain is appropriately seeded and ready for use.

    Raises AssertionError's if the brain is not ready.

    :param connection:  <rethinkdb.net.DefaultConnection>
    :param requirements:<dict> keys=Required Databases, key-values=Required Tables in each database
    :return: <rethinkdb.net.DefaultConnection> if verified
    """
    return brain_post(connection, requirements)


@wrap_self_test
def brain_post(connection, requirements=None):
    """
    Power On Self Test for the brain.

    Checks that the brain is appropriately seeded and ready for use.

    Raises AssertionError's if the brain is not ready.

    :param connection:  <rethinkdb.net.DefaultConnection>
    :param requirements:<dict> keys=Required Databases, key-values=Required Tables in each database
    :return: <rethinkdb.net.DefaultConnection> if verified
    """
    assert isinstance(connection, DefaultConnection)
    remote_dbs = validate_get_dbs(connection)
    assert validate_brain_requirements(connection, remote_dbs, requirements)
    assert validate_write_access(connection)
    return connection


def connect(host=None,
            port=rethinkdb.DEFAULT_PORT,
            timeout=20,
            verify=True,
            **kwargs):
    """
    RethinkDB semantic connection wrapper

    raises <brain.connection.BrainNotReady> if connection verification fails

    :param verify: <bool> (default True) whether to run POST
    :param timeout: <int> max time (s) to wait for connection
    :param kwargs:  <dict> passthrough rethinkdb arguments
    :return:
    """
    if not host:
        host = DEFAULT_HOSTS.get(check_stage_env())
    connection = None
    tries = 0
    time_quit = time() + timeout
    while not connection and time() <= time_quit:
        tries += 1
        connection = _attempt_connect(host, port, timeout/3, verify, **kwargs)
        if not connection:
            sleep(0.5)
    if not connection:
        raise BrainNotReady(
            "Tried ({}:{}) {} times at {} second max timeout".format(host,
                                                                     port,
                                                                     tries,
                                                                     timeout))
    return connection


def _attempt_connect(host, port, timeout, verify, **kwargs):
    """
    Internal function to attempt
    :param host: <str> "localhost" or IPAddress
    :param port: <int>
    :param timeout: <int>
    :param verify: <bool>
    :param kwargs: <**dict> rethinkdb keyword args
    :return: <connection> or <NoneType>
    """
    try:
        connection = rethinkdb.connect(host,
                                       port,
                                       timeout=timeout,
                                       **kwargs)
        if verify:
            brain_post(connection)
    except (rethinkdb.errors.ReqlDriverError,
            rethinkdb.errors.ReqlOpFailedError,
            AssertionError):
        connection = None
    return connection

#RecursiveImports
from .queries import create_plugin, destroy_plugin, RPX



"""
functions related to moving binary objects in/out of Brain.Files
"""

from .. import r
from ..queries.decorators import wrap_connection, wrap_rethink_errors
from ..queries.decorators import wrap_connection_reconnect_test
from ..checks import verify
from ..brain_pb2 import Binary
from ..queries import RBF
from .decorators import wrap_name_to_id, wrap_guess_content_type
from .decorators import wrap_content_as_binary_if_needed
from .decorators import PRIMARY_FIELD, CONTENT_FIELD

BINARY = r.binary


@wrap_rethink_errors
@wrap_name_to_id
@wrap_guess_content_type
@wrap_content_as_binary_if_needed
@wrap_connection_reconnect_test
@wrap_connection
def put(obj_dict, conn=None, **kwargs):
    """
    This function might thorw an error like:
        Query size (290114573) greater than maximum (134217727)
    caller may need to manually split files at this time
    :param obj_dict: <dict> matching brain_pb2.Binary object
    :param conn:
    :param kwargs:
    :return:
    """
    if kwargs.get("verify", False):
        verify(obj_dict, Binary())
    inserted = RBF.insert(obj_dict).run(conn)
    return inserted


def put_buffer(filename, content, conn=None):
    """
    helper function for put
    :param filename: <str>
    :param content: <bytes>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict>
    """
    obj_dict = {PRIMARY_FIELD: filename,
                CONTENT_FIELD: content}
    return put(obj_dict, conn=conn)


@wrap_rethink_errors
@wrap_connection
def get(filename, conn=None):
    """

    :param filename:
    :param conn:
    :return: <dict>
    """
    return RBF.get(filename).run(conn)


@wrap_rethink_errors
@wrap_connection
def list_dir(conn=None):
    """

    :param conn:
    :return: <list>
    """
    available = RBF.pluck(PRIMARY_FIELD).run(conn)
    return [x[PRIMARY_FIELD] for x in available]


@wrap_rethink_errors
@wrap_connection
def delete(filename, conn=None):
    """
    deletes a file
    filename being a value in the "id" key
    :param filename: <str>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict>
    """
    return RBF.get(filename).delete().run(conn)



"""
Query module is getting big
"""

from ..connection import rethinkdb as r
# recursive imports at bottom of file


RBT = r.db("Brain").table("Targets")
RBJ = r.db("Brain").table("Jobs")
RBO = r.db("Brain").table("Outputs")
RBF = r.db("Brain").table("Files")
RPX = r.db("Plugins")
RPC = r.db("Controller").table("Plugins")
RPP = r.db("Controller").table("Ports")

# backward compatible api
from .reads import *
from .writes import *
